#!/usr/bin/ruby
#
# rlisp.rb - a very simple LISP implimentation on Ruby
#
# Author:: FUJIWARA Teruyoshi <tf (at) dsl.gr.jp>
# Copyright:: Copyright (c) 2011 FUJIWARA Teruyoshi
# License::   Distributes under the same terms as Ruby
# Date:: April 20, 2011
# Version:: 1.1

# This implementation is alomst 'pure LISP', which includes just
# 5 basic functions(car, cdr, cons, atom and eq).
# Some important functions are also included(lambda, funcall, if,
# define, quote, +, -, *, <= and >=).

require 'singleton'
#
# Type system
#
module Sexp
end

class Atom
  include Sexp
end

class Sym < Atom
  def initialize(name)
    @name = name
    @value = $nil_sexp
  end
  
  def set_value(value)
    @value = value
  end
  
  def print_sexp
    print @name
  end
  
  def name
    return @name
  end
  
  def value
    return @value
  end
  
  def equal(obj)
    if (symbolp(obj))
      return (obj.name == @name)
    else
      return false
    end
  end
end

class Number < Atom
end

class Int < Number
  def initialize(i)
    @value = i
  end
  
  def print_sexp
    print "#{@value}"
  end
  
  def value
    @value
  end
  
  def equal(obj)
    if (numberp(obj))
      return (obj.value == @value)
    else
      return false
    end
  end
  
  def le(obj)
    if (numberp(obj))
      return (obj.value <= @value)
    else
      return false
    end
  end
  
  def ge(obj)
    if (numberp(obj))
      return (obj.value >= @value)
    else
      return false
    end
  end
end

class Nil < Atom
  include Singleton
  include Sexp
  
  def print_sexp
    print "nil"
  end
  
  def car
    return $nil_sexp
  end
  
  def cdr
    return $nil_sexp
  end

  def equal(obj)
    return (obj == $nil_sexp)
  end
end

class T < Atom
  include Singleton
  
  def print_sexp
    print "t"
  end
  
  def equal(obj)
    return (obj == $t_sexp)
  end
end

class List
  def initialize
    @car = nil
    @cdr = $nil_sexp
  end
  
  def car
    return @car
  end
  
  def cdr
    return @cdr
  end
  
  def set_car(v)
    @car = v
  end
  
  def set_cdr(v)
    @cdr = v
  end
  
  def equal(obj)
    return false unless (listp(obj))
    
    l1 = self
    l2 = obj
    while (l1 != $nil_sexp && l2 != $nil_sexp)
      if (!l1.car.equal(l2.car))
        return false
      end
      l1 = l1.cdr
      l2 = l2.cdr
    end
    return false unless (l1 == $nil_sexp)
    return false unless (l2 == $nil_sexp)
    return true
  end
  
  def print_sexp
    ls = self
    print "("
    while true
      if (ls == $nil_sexp)
        break
      end
      ls.car.print_sexp
      print " "
      ls = ls.cdr
    end
    print ")"
  end
end

class SymbolTable
  def initialize
    @hash = Hash.new
  end
  
  def set(name, value)
    @hash[name] = value
  end
  
  def get(name)
    if (@hash.has_key?(name))
      return @hash[name];
    else
      return $nil_sexp
    end
  end
  
  def remove(name)
    @hash.delete(name)
  end
end

class GenSym
  include Singleton
  def initialize
    @seq = 0
  end
  
  def get
    @seq = @seq + 1
    return get_symbol("<func:#{@seq}>")
  end
end

def atomp(sexp)
  return (sexp.instance_of?(Atom) || sexp.kind_of?(Atom))
end

def symbolp(sexp)
  return sexp.instance_of?(Sym)
end

def numberp(sexp)
  return sexp.kind_of?(Number)
end

def listp(sexp)
  return sexp.instance_of?(List)
end

#
# Reader
#
class Line
  def initialize(str)
    @str = str
    @idx = 0
    @prev_idx = 0
  end
  
  def unget_token
    @idx = @prev_idx
  end
  
  def get_token
    @prev_idx = @idx
    result = ""
    if (@idx == @str.length)
      return nil
    end
    while true
      ch = @str.slice(@idx, 1)
      if (ch != " ")
        break
      end
      @idx = @idx + 1
    end
    while true
      if (@idx == @str.length)
        break
      end
      ch = @str.slice(@idx, 1)
      if (ch == ")" || ch == "(" || ch == " ")
        @idx = @idx + 1
        if (result.length > 0)
          @idx = @idx - 1 if ch == ")"
          return result
        else
          return ch
        end
      end
      result = result + ch
      @idx = @idx + 1
    end
    return result
  end
end

def get_list(line)
  t = line.get_token
  if (t == ")")
    return $nil_sexp
  end
  top = List.new
  ls = top
  line.unget_token
  ls.set_car(get_sexp(line))
  
  while true
    t = line.get_token
    if (t == nil)
      break
    end
    if (t == ")")
      return top
    end
    new_cell = List.new
    line.unget_token
    new_cell.set_car(get_sexp(line))
    ls.set_cdr(new_cell)
    ls = new_cell
  end
  return $nil_sexp
end

def get_number(token)
  return Int.new(token.to_i)
end

def get_symbol(token)
  return Sym.new(token)
end

def get_sexp(l)
  while t = l.get_token
    if (t == "(")
      return get_list(l)
    elsif (t =~ /^[0-9]+$/)
      return get_number(t)
    elsif (t == "t")
      return $t_sexp
    elsif (t == "nil")
      return $nil_sexp
    else
      return get_symbol(t)
    end
  end
  return $nil_sexp
end

#
# Embedded functions
#

# Pure lisp functions (car, cdr, cons, atom, eq)
def func_car(ls)
  if (listp(ls))
    arg = eval_sexp(ls.car)
    return eval_sexp(arg.car)
  else
    print "Invalid argument: car\n"
  end
end

def func_cdr(ls)
  if (listp(ls))
    arg = eval_sexp(ls.car)
    return arg.cdr
  else
    print "Invalid argument: cdr\n"
  end
end

def func_cons(ls)
  car = eval_sexp(ls.car)
  cadr = eval_sexp(ls.cdr.car)
  ln = List.new
  ln.set_car(car)
  ln.set_cdr(cadr)
  return ln
end

def func_atom(ls)
  if (atomp(ls.car))
    return $t_sexp
  else
    return $nil_sexp
  end
end

def func_eq(ls)
  ref = eval_sexp(ls.car)
  cdr = ls.cdr
  while (cdr != $nil_sexp)
    val = eval_sexp(cdr.car)
    unless (ref.equal(val))
      return $nil_sexp
    end
    cdr = cdr.cdr
  end
  return $t_sexp
end

# additional functions
def func_lambda(ls)
  sym = $gensym.get
  $symtable.set(sym.name, ls)
  return sym
end

def func_funcall(ls)
  return funcall(ls)
end

def func_if(ls)
  cond = ls.car
  tform = ls.cdr.car
  if (ls.cdr.cdr == $nil_sexp)
    fform = $nil_sexp
  else
    fform = ls.cdr.cdr.car
  end
  if (eval_sexp(cond) == $t_sexp)
    return eval_sexp(tform)
  else
    return eval_sexp(fform)
  end
end

def func_define(ls)
  sym = ls.car
  val = ls.cdr.car
  if (atomp(val))
    $symtable.set(sym.name, val)
    return val
  else
    $symtable.set(sym.name, eval_sexp(val))
    return sym
  end
end

def func_quote(ls)
  if (listp(ls))
    return ls
  else
    print "Invalid argument: quote\n"
  end
end

def func_plus(ls)
  if (ls == $nil_sexp)
    return Int.new(0)
  else
    return Int.new(eval_sexp(ls.car).value + func_plus(ls.cdr).value)
  end
end

def func_minus(ls)
  if (ls == $nil_sexp)
    return Int.new(0)
  else
    return Int.new(eval_sexp(ls.car).value - func_plus(ls.cdr).value)
  end
end

def func_multiply(ls)
  if (ls == $nil_sexp)
    return Int.new(1)
  else
    return Int.new(eval_sexp(ls.car).value * func_multiply(ls.cdr).value)
  end
end

def func_ge(ls)
  ref = eval_sexp(ls.car)
  cdr = ls.cdr
  while (cdr != $nil_sexp)
    val = eval_sexp(cdr.car)
    unless (ref.ge(val))
      return $nil_sexp
    end
    cdr = cdr.cdr
  end
  return $t_sexp
end

def func_le(ls)
  ref = eval_sexp(ls.car)
  cdr = ls.cdr
  while (cdr != $nil_sexp)
    val = eval_sexp(cdr.car)
    unless (ref.le(val))
      return $nil_sexp
    end
    cdr = cdr.cdr
  end
  return $t_sexp
end

#
# Evaluater
#
def funcall(ls)
  car = ls.car
  cdr = ls.cdr
  
  if (symbolp(car))
    func = $symtable.get(car.name)
    if (func == $nil_sexp)
      print "#{car} is not a function.\n"
      return $nil_sexp
    elsif (symbolp(func) && func.name =~ /^<func:/)
      lmdls = $symtable.get(func.name)
      vars = lmdls.car
      body = lmdls.cdr.car
      return eval_func(vars, body, cdr)
    else
      return func.call(cdr)
    end
  elsif (listp(car))
    func = eval_sexp(car)
    if (symbolp(func) && func.name =~ /^<func:/)
      lmdls = $symtable.get(func.name)
      vars = lmdls.car
      body = lmdls.cdr.car
      return eval_func(vars, body, cdr)
    else
      print "#{car} is not a function.\n"
      return $nil_sexp
    end
  else
    print "#{car} is not a function.\n"
    return $nil_sexp
  end
end

def eval_func(vars, body, args)
  env = Hash.new
  l = vars
  a = args
  
  while (l != $nil_sexp || a != $nil_sexp)
    key = l.car.name
    val = eval_sexp(a.car)
    env[key] = $symtable.get(key)
    $symtable.set(key, val)
    l = l.cdr
    a = a.cdr
  end
  ret = eval_sexp(body)
  env.each_pair do |key, value|
    $symtable.set(key, value)
  end
  return ret
end

def eval_sexp(sexp)
  if (numberp(sexp))
    return sexp
  elsif (sexp == $t_sexp || sexp == $nil_sexp)
    return sexp
  elsif (symbolp(sexp))
    return $symtable.get(sexp.name)
  elsif (listp(sexp))
    return funcall(sexp)
  else
    print "unexpected error.\n"
    sexp.print_sexp
  end
end

#
# main part
#
print "initializing..."
$t_sexp = T.instance
$nil_sexp = Nil.instance
$gensym = GenSym.instance

$symtable = SymbolTable.new

$symtable.set("car", self.method(:func_car))
$symtable.set("cdr", self.method(:func_cdr))
$symtable.set("cons", self.method(:func_cons))
$symtable.set("atom", self.method(:func_atom))
$symtable.set("eq", self.method(:func_eq))

$symtable.set("lambda", self.method(:func_lambda))
$symtable.set("funcall", self.method(:func_funcall))
$symtable.set("if", self.method(:func_if))
$symtable.set("define", self.method(:func_define))
$symtable.set("quote", self.method(:func_quote))

$symtable.set("+", self.method(:func_plus))
$symtable.set("-", self.method(:func_minus))
$symtable.set("*", self.method(:func_multiply))

$symtable.set("<=", self.method(:func_ge))
$symtable.set(">=", self.method(:func_le))

f = open("init.l")
while l = f.gets
  next if l =~ /^;/
  l = l.chomp
  line = Line.new(l)
  s = get_sexp(line)
  eval_sexp(s)
end
f.close
print "done.\n"

# interactive session
while l = gets
  next if l =~ /^;/
  l = l.chomp
  line = Line.new(l)
  s = get_sexp(line)
  break if (symbolp(s) && s.name == "quit")
  v = eval_sexp(s)
  print "=> "
  if (listp(v))
    v.print_sexp
  elsif (numberp(v))
    print "#{v.value}"
  elsif (v == $nil_sexp)
    print "nil"
  elsif (v == $t_sexp)
    print "t"
  elsif (symbolp(v))
    print v.name
  else
    print v.print_sexp
  end
  print "\n"
end
